/*
 * Copyright (c) 1999 Greg Haerr <greg@censoft.com>
 *
 * Windows BMP to Micro-Win image converter
 *
 * 6/9/1999 g haerr
 */
#include "../device.h"
#include <stdio.h>
#include <malloc.h>

#define PIX2BYTES(n)	(((n)+7)/8)

#define BI_RGB		0L
#define BI_RLE8		1L
#define BI_RLE4		2L
#define BI_BITFIELDS	3L

typedef unsigned char	BYTE;
typedef unsigned short	WORD;
typedef unsigned long	DWORD;
typedef long		LONG;

#pragma pack(1)
/* windows style*/
typedef struct {
	/* BITMAPFILEHEADER*/
	BYTE	bfType[2];
	DWORD	bfSize;
	WORD	bfReserved1;
	WORD	bfReserved2;
	DWORD	bfOffBits;
	/* BITMAPINFOHEADER*/
	DWORD	BiSize;
	LONG	BiWidth;
	LONG	BiHeight;
	WORD	BiPlanes;
	WORD	BiBitCount;
	DWORD	BiCompression;
	DWORD	BiSizeImage;
	LONG	BiXpelsPerMeter;
	LONG	BiYpelsPerMeter;
	DWORD	BiClrUsed;
	DWORD	BiClrImportant;
} BMPHEAD;

/* os/2 style*/
typedef struct {
	/* BITMAPFILEHEADER*/
	BYTE	bfType[2];
	DWORD	bfSize;
	WORD	bfReserved1;
	WORD	bfReserved2;
	DWORD	bfOffBits;
	/* BITMAPCOREHEADER*/
	DWORD	bcSize;
	short	bcWidth;
	short	bcHeight;
	WORD	bcPlanes;
	WORD	bcBitCount;
} BMPCOREHEAD;
#pragma pack()

BOOL		ConvBMP(FILE *fp,char *name);
static void	outline(UCHAR *linebuffer, int bitdepth, int linesize, int y);
static int	DecodeRLE8(UCHAR *buf,FILE *fp);
static int	DecodeRLE4(UCHAR *buf,FILE *fp);
static void	put4(int b);

int
main(int ac,char **av)
{
	FILE	*fp;
	char 	*p, *q;
	char	name[64];

	if(ac < 2) {
		fprintf(stderr, "Usage: convbmp <bmpfile>\n");
		exit(1);
	}
	fp = fopen(av[1], "rb");
	if(!fp) {
		fprintf(stderr, "Can't open file: %s\n", av[1]);
		exit(1);
	}
	p = av[1];
	q = name;
	while(*p && *p != '.')
		*q++ = *p++;
	*q = 0;
	if(!ConvBMP(fp, name)) {
		fprintf(stderr, "Conversion failed: %s\n", av[1]);
		exit(1);
	}
	return 0;
}

/* decode a bmp file*/
/* close input file before returning*/
BOOL
ConvBMP(FILE *fp,char *name)
{
	BMPHEAD		bmp;
	BMPCOREHEAD *	bmc;
	int		i, n;
	UCHAR *		linebuffer = NULL;
	int		cx, cy, bitdepth, linesize, compression;
	RGBENTRY	cmap[256];

	/* read BMP header*/
	if(fread(&bmp, 1, sizeof(BMPHEAD), fp) != sizeof(BMPHEAD))
		goto err;

	/* might be windows or os/2 header*/
	if(bmp.BiSize == 12) {
		bmc = (BMPCOREHEAD *)&bmp;
		cx = (int)bmc->bcWidth;
		cy = (int)bmc->bcHeight;
		bitdepth = bmc->bcBitCount;
		n = 1 << bitdepth;
		compression = BI_RGB;
		fseek(fp, sizeof(BMPCOREHEAD), SEEK_SET);
	} else {
		cx = (int)bmp.BiWidth;
		cy = (int)bmp.BiHeight;
		bitdepth = bmp.BiBitCount;
		n = (int)bmp.BiClrUsed;
		if(n == 0)
			n = 1 << bitdepth;
		compression = bmp.BiCompression;
	}

	/* currently only decode palettized bitmaps*/
	if(bitdepth > 8) {
		fprintf(stderr, "Error: bits per pixel > 8\n");
		goto err;
	}

	/* compute image line size and allocate line buffer*/
	if(bitdepth == 1)
		linesize = PIX2BYTES(cx);
	else if(bitdepth <= 4)
		linesize = PIX2BYTES(cx<<2);
	else if(bitdepth <= 8)
		linesize = cx;
	else if(bitdepth <= 16)
		linesize = cx * 2;
	else if(bitdepth <= 24)
		linesize = cx * 3;
	else
		linesize = cx * 4;
	linesize = (linesize+3) & ~3;

	if( (linebuffer = malloc(linesize)) == NULL)
		goto err;

	/* get colormap*/
	if(bitdepth <= 8) {
		for(i=0; i<n; i++) {
			cmap[i].b = fgetc(fp);
			cmap[i].g = fgetc(fp);
			cmap[i].r = fgetc(fp);
			if(bmp.BiSize != 12)
				fgetc(fp);
		}
	}

	printf("/* Generated by convbmp*/\n");
	printf("#include \"device.h\"\n\n");
	printf("/* IMAGEHDR image_%s converted from %s.bmp*/\n\n",
		name, name);

	/* extract palette*/
	printf("static RGBENTRY palette[%d] = {\n", n);
	for(i=0; i<n; ++i)
		printf("  RGBDEF( %3d, %3d, %3d ),\t/* pal %d*/\n",
			cmap[i].r, cmap[i].g, cmap[i].b, i);
	printf("};\n\n");

	/* decode image data*/
	printf("static UCHAR imagebits[] = {\n", n);

	fseek(fp, bmp.bfOffBits, SEEK_SET);
	if(compression == BI_RLE8) {
		for(i = cy-1; i>= 0; i--) {
			if(!DecodeRLE8(linebuffer, fp))
				break;
			outline(linebuffer, bitdepth, linesize, i);
		}
	} else if(compression == BI_RLE4) {
		for(i = cy-1; i>= 0; i--) {
			if(!DecodeRLE4(linebuffer, fp))
				break;
			outline(linebuffer, bitdepth, linesize, i);
		}
	} else {
		for(i=0; i<cy; i++) {
			if(fread(linebuffer, 1, linesize, fp) != (size_t)linesize)
				goto err;
			outline(linebuffer, bitdepth, linesize, cy-i-1);
		}
	}
	printf("};\n\n");

	printf("IMAGEHDR image_%s = {\n", name);
	printf("  %d, %d,\t/* width, height*/\n", cx, cy);
	printf("  %d, %d,\t\t/* planes, bpp*/\n", 1, bitdepth);
	printf("  %d, %d,\t/* compression, palsize*/\n", 1, n);
	printf("  palette,\n");
	printf("  imagebits,\n");
	printf("};\n");

	free(linebuffer);
	fclose(fp);
	return TRUE;

err:
	if(linebuffer)
		free(linebuffer);
	fclose(fp);
	return FALSE;
}

static void
outline(UCHAR *linebuffer, int bitdepth, int linesize, int y)
{
	int	n;

	switch(bitdepth) {
	case 1:
	case 4:
		for(n=0; n<linesize; ++n)
			printf("0x%02x,", linebuffer[n]);
		break;
	case 8:
	default:
		for(n=0; n<linesize; ++n)
			printf("%d,", linebuffer[n]);
	}
	printf("\n");
}

/*
 * Decode one line of RLE8, return 0 when done with all bitmap data
 */
static int
DecodeRLE8(UCHAR *buf,FILE *fp)
{
	int		c, n;
	UCHAR *	p = buf;

	for( ;;) {
		switch( n = fgetc( fp)) {
		case EOF:
			return( 0);
		case 0:					/* 0 = escape*/
			switch( n = fgetc( fp)) {
			case 0: 			/* 0 0 = end of current scan line*/
				return( 1);
			case 1:				/* 0 1 = end of data*/
				return( 1);
			case 2:				/* 0 2 xx yy delta mode - NOT SUPPORTED*/
				(void)fgetc( fp);
				(void)fgetc( fp);
				continue;
			default:			/* 0 3..255 xx nn uncompressed data*/
				for( c=0; c<n; c++)
					*p++ = fgetc( fp);
				if( n & 1)
					(void)fgetc( fp);
				continue;
			}
		default:
			c = fgetc( fp);
			while( n--)
				*p++ = c;
			continue;
		}
	}
}

/*
 * Decode one line of RLE4, return 0 when done with all bitmap data
 */
static UCHAR *p;
static int	once;

static void
put4(int b)
{
	static int	last;

	last = (last << 4) | b;
	if( ++once == 2) {
		*p++ = last;
		once = 0;
	}
}
	
static int
DecodeRLE4(UCHAR *buf,FILE *fp)
{
	int		c, n, c1, c2;

	p = buf;
	once = 0;
	for( ;;) {
		switch( n = fgetc( fp)) {
		case EOF:
			return( 0);
		case 0:					/* 0 = escape*/
			switch( n = fgetc( fp)) {
			case 0: 			/* 0 0 = end of current scan line*/
				if( once)
					put4( 0);
				return( 1);
			case 1:				/* 0 1 = end of data*/
				if( once)
					put4( 0);
				return( 1);
			case 2:				/* 0 2 xx yy delta mode - NOT SUPPORTED*/
				(void)fgetc( fp);
				(void)fgetc( fp);
				continue;
			default:			/* 0 3..255 xx nn uncompressed data*/
				c2 = (n+3) & ~3;
				for( c=0; c<c2; c++) {
					if( (c & 1) == 0)
							c1 = fgetc( fp);
					if( c < n)
							put4( (c1 >> 4) & 0x0f);
					c1 <<= 4;
				}
				continue;
			}
		default:
			c = fgetc( fp);
			c1 = (c >> 4) & 0x0f;
			c2 = c & 0x0f;
			for( c=0; c<n; c++)
				put4( (c&1)? c2: c1);
			continue;
		}
	}
}
